#!/usr/bin/env python
# coding: utf-8
"""
Main functions of pangres that will be directly exposed to its users.
"""
import warnings
from pangres.helpers import PandasSpecialEngine


# # upsert

def upsert(engine,
           df,
           table_name,
           if_row_exists,
           schema=None,
           create_schema=False,
           add_new_columns=False,
           adapt_dtype_of_empty_db_columns=False,
           chunksize=10000,
           dtype=None):
    """
    Insert updates/ignores a pandas DataFrame into a SQL table (or
    creates a SQL table from the DataFrame if it does not exist).

    This index of the DataFrame must be the primary key or a unique key
    (can be a unique constraint with multiple columns too) of the SQL
    table and meet the following conditions:
    1. The index must be unique.
    2. All levels of the index have to be named.
    3. The index cannot contain any null value except for MySQL where
    rows with null in indices will be skipped and warnings will be raised.

    **GOTCHAS**:

    Please head over to https://github.com/ThibTrip/pangres/#Gotchas-and-caveats
    or read pangres/README.md.

    Notes
    -----
    It is recommanded to use this function with big batches of data
    as there is quite the overhead. Setting the arguments create_schema,
    add_new_columns and adapt_dtype_of_empty_db_columns to False should
    drastically reduce the overhead if you do not need such features.

    Parameters
    ----------
    engine : sqlalchemy.engine.base.Engine
        See https://docs.sqlalchemy.org/en/latest/core/engines.html
    df : pd.DataFrame
        See https://pandas.pydata.org/pandas-docs/stable/getting_started/10min.html
    table_name : str
        Name of the SQL table
    if_row_exists : {'ignore', 'update'}
        Behavior if a row exists in the SQL table.
        Irrelevant if the SQL table does not exist already.
        For both 'ignore' and 'update' rows with new primary/unique
        key values are inserted.

        * 'ignore': rows for which primary/unique keys exist in the SQL
        table are skipped
        * 'update': rows for which primary/unique keys exist in the SQL
        table are updated (with the columns available in the pandas
        DataFrame)
    schema : str or None, default None
        Name of the SQL schema that contains or will contain the table
        For postgres if it is None it will default to "public".
        For MySQL and SQlite the schema should be None since
        those SQL flavors do not have this system.
    create_schema : bool, default False
        If True the schema is created if it does not exist
    add_new_columns : bool, default False
        If True adds columns present in the DataFrame that
        are not in the SQL table.
    adapt_dtype_of_empty_db_columns : bool, default False
        If True looks for columns that have no data in the
        SQL table but have data in the DataFrame;
        if those columns have datatypes that do not match
        (e.g. "TEXT" in the SQL table and "int64" in the DataFrame)
        then they are altered in the SQL table.
        Data type conversion must be supported by the SQL flavor!
        E.g. for Postgres converting from BOOLEAN to TIMESTAMP
        will not work even if the column is empty.
    chunksize : int
        Number of rows to insert at once.
        Please note that for SQlite a maximum of 999 parameters
        per queries means that the chunksize will be automatically
        reduced to math.floor(999/nb_columns) where nb_columns is
        the number of columns + index levels in the DataFrame.
    dtype : None or dict {str:SQL_TYPE}, default None
        Similar to pd.to_sql dtype argument.
        This is especially useful for MySQL where the length of
        primary keys with text has to be provided (see Examples)

    Examples
    --------
    #### 1. Workflow example

    ##### 1.1. Creating a SQL table
    >>> import pandas as pd
    >>> from pangres import upsert, DocsExampleTable
    >>> from sqlalchemy import create_engine, VARCHAR
    >>> 
    >>> # create a SQLalchemy engine
    >>> engine = create_engine("sqlite:///:memory:")
    >>> 
    >>> # this is necessary if you want to test with MySQL
    >>> # instead of SQlite or Postgres because MySQL needs
    >>> # to have a definite limit for text primary keys/indices
    >>> dtype = {'full_name':VARCHAR(50)}
    >>> 
    >>> # get or create a pandas DataFrame
    >>> # for our example full_name is the index
    >>> # and will thus be used as primary key
    >>> df = DocsExampleTable.df
    >>> print(df.to_markdown()) # to_markdown exists since pandas v1
    | full_name     | likes_sport   | updated                   |   size_in_meters |
    |:--------------|:--------------|:--------------------------|-----------------:|
    | John Rambo    | True          | 2020-02-01 00:00:00+00:00 |             1.77 |
    | The Rock      | True          | 2020-04-01 00:00:00+00:00 |             1.96 |
    | John Travolta | False         | NaT                       |           nan    |

    >>> # create SQL table
    >>> # it does not matter if if_row_exists is set
    >>> # to "update" or "ignore" for table creation
    >>> upsert(engine=engine,
    ...        df=df,
    ...        table_name='example',
    ...        if_row_exists='update',
    ...        dtype=dtype)

    ##### 1.2. Updating the SQL table we created with if_row_exists='update'
    >>> new_df = DocsExampleTable.new_df
    >>> print(new_df.to_markdown())
    | full_name             | likes_sport   | updated                   |   size_in_meters |
    |:----------------------|:--------------|:--------------------------|-----------------:|
    | John Travolta         | True          | 2020-04-04 00:00:00+00:00 |             1.88 |
    | Arnold Schwarzenegger | True          | NaT                       |             1.88 |

    >>> # insert update using our new data
    >>> upsert(engine=engine,
    ...        df=new_df,
    ...        table_name='example',
    ...        if_row_exists='update',
    ...        dtype=dtype)
    >>> 
    >>> # Now we read from the database to check what we got and as you can see
    >>> # John Travolta was updated and Arnold Schwarzenegger was added!
    >>> print(pd.read_sql('SELECT * FROM example', con=engine, index_col='full_name')
    ...  .astype({'likes_sport':bool}).to_markdown())
    | full_name             | likes_sport   | updated                    |   size_in_meters |
    |:----------------------|:--------------|:---------------------------|-----------------:|
    | John Rambo            | True          | 2020-02-01 00:00:00.000000 |             1.77 |
    | The Rock              | True          | 2020-04-01 00:00:00.000000 |             1.96 |
    | John Travolta         | True          | 2020-04-04 00:00:00.000000 |             1.88 |
    | Arnold Schwarzenegger | True          |                            |             1.88 |

    ##### 1.3. Updating the SQL table with if_row_exists='ignore'
    >>> new_df2 = DocsExampleTable.new_df2
    >>> print(new_df2.to_markdown())
    | full_name     | likes_sport   | updated   |   size_in_meters |
    |:--------------|:--------------|:----------|-----------------:|
    | John Travolta | True          | NaT       |             2.5  |
    | John Cena     | True          | NaT       |             1.84 |

    >>> upsert(engine=engine,
    ...        df=new_df2,
    ...        table_name='example',
    ...        if_row_exists='ignore',
    ...        dtype=dtype)
    >>> # Now we read from the database to check what we got and as you can see
    >>> # John Travolta was NOT updated and John Cena was added!
    >>> print(pd.read_sql('SELECT * FROM example', con=engine, index_col='full_name')
    ...  .astype({'likes_sport':bool}).to_markdown())
    | full_name             | likes_sport   | updated                    |   size_in_meters |
    |:----------------------|:--------------|:---------------------------|-----------------:|
    | John Rambo            | True          | 2020-02-01 00:00:00.000000 |             1.77 |
    | The Rock              | True          | 2020-04-01 00:00:00.000000 |             1.96 |
    | John Travolta         | True          | 2020-04-04 00:00:00.000000 |             1.88 |
    | Arnold Schwarzenegger | True          |                            |             1.88 |
    | John Cena             | True          |                            |             1.84 |
    """
    pse = PandasSpecialEngine(engine=engine,
                              df=df,
                              table_name=table_name,
                              schema=schema,
                              dtype=dtype)

    # add new columns from frame
    if add_new_columns and pse.table_exists():
        pse.add_new_columns()

    # change dtype of empty columns in db
    if adapt_dtype_of_empty_db_columns and pse.table_exists():
        pse.adapt_dtype_of_empty_db_columns()

    # create schema and table if not exists then insert values
    if create_schema and schema is not None:
        pse.create_schema_if_not_exists()
    pse.create_table_if_not_exists()
    
    # stop if no rows
    if df.empty:
        return
    pse.upsert(if_row_exists=if_row_exists, chunksize=chunksize)



# +
import pandas as pd
import logging

# configure logger
logging_format = ('%(asctime)s | %(levelname)s     '
                  '| pangres     | %(module)s:%(funcName)s:%(lineno)s '
                  '- %(message)s')
logging.basicConfig(format=logging_format, datefmt='%Y-%m-%d %H:%M:%S')
logger = logging.getLogger('pangres')
logger.setLevel(logging.INFO)


# -

# # Function to fix bad column names for psycopg2

def fix_psycopg2_bad_cols(df:pd.DataFrame, replacements={'%':'', '(':'', ')':''}) -> pd.DataFrame:
    """
    Replaces '%', '(' and ')' (characters that won't play nicely or even
    at all with psycopg2) in column and index names in a deep copy of df.
    This is a workaround for the unresolved issue
    described here: https://github.com/psycopg/psycopg2/issues/167

    **IMPORTANT**:
    You will need to apply the same changes in the database as
    well if the SQL table already exists for a given DataFrame.
    Otherwise you will for instance end up with a column
    "total_%" and "total_" in your SQL table.

    Parameters
    ----------
    df : pd.DataFrame
    replacements : dict {'%':str, '(':str, ')':str}, default {'%':'', '(':'', ')':''}
        The keys '%', '(' and ')' are mandatory.
        There cannot be any extra keys.

    Returns
    -------
    new_df : pd.DataFrame

    Examples
    --------
    * fix bad col/index names with default replacements (empty string for '(', ')' and '%')
    >>> import pandas as pd
    >>> df = pd.DataFrame({'test()':[0],
    ...                    'foo()%':[0]}).set_index('test()')
    >>> print(df.to_markdown())
    |   test() |   foo()% |
    |---------:|---------:|
    |        0 |        0 |

    >>> df_fixed = fix_psycopg2_bad_cols(df)
    >>> print(df_fixed.to_markdown())
    |   test |   foo |
    |-------:|------:|
    |      0 |     0 |

    * fix bad col/index names with custom replacements - you MUST provide replacements for '(', ')' and '%'!
    >>> import pandas as pd
    >>> df = pd.DataFrame({'test()':[0],
    ...                    'foo()%':[0]}).set_index('test()')
    >>> print(df.to_markdown())
    |   test() |   foo()% |
    |---------:|---------:|
    |        0 |        0 |

    >>> df_fixed = fix_psycopg2_bad_cols(df, replacements={'%':'percent', '(':'', ')':''})
    >>> print(df_fixed.to_markdown())
    |   test |   foopercent |
    |-------:|-------------:|
    |      0 |            0 |
    """
    # verify all index levels are named
    index_names = list(df.index.names)
    if any([ix_name is None for ix_name in index_names]):
        raise IndexError("All index levels must be named!")
    # verify duplicated columns
    fields = list(df.index.names) + df.columns.tolist()
    if len(set(fields)) != len(fields):
        raise ValueError(("There cannot be duplicated names amongst "
                          "index levels and/or columns!"))
    # verify replacements arg
    expected_keys = ('%', '(', ')')
    if ((not isinstance(replacements, dict)) or
        (set(replacements.keys()) - set(expected_keys) != set()) or 
        (len(replacements) != len(expected_keys))):
        raise TypeError(f'replacements must be a dict containing the following keys (and none other): {expected_keys}')
    if not all((isinstance(v, str) for v in replacements.values())):
        raise TypeError(f'The values of replacements must all be strings')
    # replace bad col names
    translator = {ord(k):v for k, v in replacements.items()}
    new_df = df.copy(deep=True)
    renamer = lambda col: col.translate(translator) if isinstance(col, str) else col
    new_df = new_df.rename(columns=renamer).rename_axis(index=renamer)
    # check columns are unique after renaming
    fields = list(new_df.index.names) + new_df.columns.tolist()
    if len(set(fields)) != len(fields):
        raise ValueError("Columns/index are not unique after renaming!")
    # compare columns (and index)
    before = df.reset_index().columns.tolist()
    after = new_df.reset_index().columns.tolist()
    changed = []
    for i, j in zip(before, after):
        if (isinstance(i, str) and isinstance(j, str)
            and i!=j):
            logger.info(f'Renamed column/index "{i}" to "{j}s"')
    return new_df


# # Test with doctest

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# coding: utf-8
import pandas as pd
import pytest
from sqlalchemy import VARCHAR
from pangres.examples import _TestsExampleTable
from pangres.helpers import PandasSpecialEngine
from pangres.tests.conftest import drop_table_if_exists


# # Test methods and attributes

def test_methods_and_attributes(engine, schema):
    """
    Makes tests in chain for PandasSpecialEngine.
    This makes code easier with a small inconvenient
    of not separating tests thematically.
    """
    # dtype for index for MySQL... (can't have flexible text length)
    dtype = {'profileid':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    table_name = 'test_pandas_special_engine'
    default_args = {'engine':engine,
                    'schema':schema,
                    'dtype':dtype,
                    'table_name':table_name}
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    # TEST INIT
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    pse = PandasSpecialEngine(df=df, **default_args)
    # TEST ATTRIBUTE pse.table
    expected_cols = list(df.index.names) + df.columns.tolist()
    assert all((col in pse.table.columns for col in expected_cols))  
    # TEST TABLE AND SCHEMA CREATION
    pse.create_schema_if_not_exists()
    pse.create_table_if_not_exists()

    # TEST ADD NEW COLUMNS
    # don't try to add JSON columns!
    # It's not supported by sqlalchemy compilers :(
    df = df.assign(new_text_col='test',
                   new_int_col=0,
                   new_float_col=1.1,
                   new_bool_col=False,
                   new_dt_col=pd.Timestamp('2020-01-01'),
                   # create this col for later
                   empty_col=None)
    # recreate pse then add columns
    pse = PandasSpecialEngine(df=df, **default_args)
    pse.add_new_columns()
    
    # TEST CHANGE COLUMN TYPE (not for SQlite)
    if not pse._db_type == 'sqlite':
        # don't try to alter from any type to JSON!
        # It's not supported by sqlalchemy compilers :(
        # also the order is very specific!!! we have to cast types
        # even though the column is empty (to avoid losing column information such as constraint)
        # so e.g. casting from BOOLEAN to BIGINT is not possible
        # actually BOOLEAN is not even here because it breaks the chain
        alterations = (1, 1.1, "abc", pd.Timestamp("2020-01-01", tz='UTC'))
        for i in alterations:
            # change empty_col
            df['empty_col'] = df['empty_col'].map(lambda x: i) # this will work for lists or dicts as well
            # recreate pse then change column type
            pse = PandasSpecialEngine(df=df, **default_args)
            pse.adapt_dtype_of_empty_db_columns()


# # Test errors

def test_errors(engine, schema):
    """
    Test errors raised by PandasSpecialEngine methods
    when possible.
    """
    table_name = 'test_pandas_special_engine_errors'
    default_args = {'engine':engine,
                    'schema':schema,
                    # this is most likely irrelevant (no table shall be created)
                    'table_name':table_name}
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    # no index name
    df = pd.DataFrame({'test':[1]})
    with pytest.raises(IndexError) as excinfo:
        pse = PandasSpecialEngine(df=df, **default_args)
    assert "All index levels must be named" in str(excinfo.value)

    # duplicates names amonst columns/index levels
    df = pd.DataFrame({'test':[1]})
    df.index.name = 'test'
    with pytest.raises(ValueError) as excinfo:
        pse = PandasSpecialEngine(df=df, **default_args)
    assert "There cannot be duplicated names" in str(excinfo.value)

    # non unique index
    df = pd.DataFrame({'test':[1, 2]}).assign(index=[0,0]).set_index('index')
    with pytest.raises(IndexError) as excinfo:
        pse = PandasSpecialEngine(df=df, **default_args)
    assert "The index must be unique" in str(excinfo.value)

    # provide dumb chunksize in
    # PandasSpecialEngine._get_values_to_insert
    # create a dummy PandasSpecialEngine
    df = pd.DataFrame({'test':[1]})
    df.index.name = 'profileid'
    pse = PandasSpecialEngine(df=df, **default_args)
    # chunksize should be an integer above 0
    for dumb_value in (0, -1, 1.2):
        with pytest.raises(ValueError) as excinfo:
            pse._create_chunks(values=[0], chunksize=dumb_value)
        assert "integer strictly above 0" in str(excinfo.value)



# +
"""
Functions/classes/variables for interacting between a pandas DataFrame
and postgres/mysql/sqlite (and potentially other databases).
"""
import json
import pandas as pd
import logging
import re
from copy import deepcopy
from math import floor
from sqlalchemy import JSON, MetaData, select
from sqlalchemy.sql import null
from sqlalchemy.schema import (PrimaryKeyConstraint, CreateColumn, CreateSchema)
from alembic.runtime.migration import MigrationContext
from alembic.operations import Operations
from pangres.upsert import (mysql_upsert,
                            postgres_upsert,
                            sqlite_upsert)

# configure logger
logging_format = ('%(asctime)s | %(levelname)s     '
                  '| pangres     | %(module)s:%(funcName)s:%(lineno)s '
                  '- %(message)s')
logging.basicConfig(format=logging_format, datefmt='%Y-%m-%d %H:%M:%S')
logger = logging.getLogger('pangres')
logger.setLevel(logging.INFO)

# compile some regexes
# column names that will cause issues with psycopg2 default parameter style
# (so we will need to switch to format style when we see such columns)
RE_BAD_COL_NAME = re.compile('[\(\)\%]')
# e.g. match "(50)" in "VARCHAR(50)"
RE_CHARCOUNT_COL_TYPE = re.compile('(?<=.)+\(\d+\)')


# -

# # Class PandasSpecialEngine

class PandasSpecialEngine:

    def __init__(self,
                 engine,
                 df,
                 table_name,
                 schema=None,
                 dtype=None):
        """
        Interacts with SQL tables via pandas and SQLalchemy table models.

        Attributes
        ----------
        engine : sqlalchemy.engine.base.Engine
            Engine provided during class instantiation
        df : pd.DataFrame
            DataFrame provided during class instantiation
        table_name : str
            Table name provided during class instantiation
        schema : str or None
            SQL schema provided during class instantiation
        table : sqlalchemy.sql.schema.Table
            Sqlalchemy table model for df

        Parameters
        ----------
        engine : sqlalchemy.engine.base.Engine
            Engine from sqlalchemy (see https://docs.sqlalchemy.org/en/13/core/engines.html
            and examples below)
        df : pd.DataFrame
            A pandas DataFrame
        table_name : str
            Name of the SQL table
        schema : str or None, default None
            Name of the schema that contains/will contain the table
            For postgres defaults to "public" if not provided.
        dtype : None or dict {str:SQL_TYPE}, default None
            Similar to pd.to_sql dtype argument.
            This is especially useful for MySQL where the length of
            primary keys with text has to be provided (see Examples)
        
        Examples
        --------
        >>> from sqlalchemy import create_engine
        >>> from pangres.helpers import PandasSpecialEngine
        >>> 
        >>> engine = create_engine("postgresql://user:password@host.com:5432/database")
        >>> df = pd.DataFrame({'name':['Albert', 'Toto'],
        ...                    'profileid':[10, 11]}).set_index('profileid')
        >>> pse = PandasSpecialEngine(engine=engine, df=df, table_name='example')
        >>> pse # doctest: +SKIP
        PandasSpecialEngine (id 123546, hexid 0x1E29A)
        * connection: Engine(postgresql://user:***@host.com:5432/database)
        * schema: public
        * table: example
        * SQLalchemy table model:
        Table('example', MetaData(bind=Engine(postgresql://user:***@host.com:5432/database)),
              Column('profileid', BigInteger(), table=<example>, primary_key=True, nullable=False),
              Column('name', Text(), table=<example>), schema='public')
        * df.head():
        |   profileid | name   |
        |------------:|:-------|
        |          10 | Albert |
        |          11 | Toto   |
        """
        self._db_type = self._detect_db_type(engine)
        if self._db_type == "postgres":
            schema = 'public' if schema is None else schema
            # raise if we find columns with "(", ")" or "%"
            if any((RE_BAD_COL_NAME.search(col) for col in df.columns)):
                err = ("psycopg2 (Python postgres driver) does not seem to support"
                       "column names with '%', '(' or ')' "
                       "(see https://github.com/psycopg/psycopg2/issues/167)")
                raise ValueError(err)

        # VERIFY ARGUMENTS
        # all index levels have names
        index_names = list(df.index.names)
        if any([ix_name is None for ix_name in index_names]):
            raise IndexError("All index levels must be named!")

        # index is unique
        if not df.index.is_unique:
            err = ("The index must be unique since it is used "
                   "as primary key.\n"
                   "Check duplicates using this code (assuming df "
                   " is the DataFrame you want to upsert):\n"
                   ">>> df.index[df.index.duplicated(keep=False)]")
            raise IndexError(err)

        # there are no duplicated names
        fields = list(df.index.names) + df.columns.tolist()
        if len(set(fields)) != len(fields):
            raise ValueError(("There cannot be duplicated names amongst "
                              "index levels and/or columns!"))

        # detect json columns
        def is_json(col):
            s = df[col].dropna()
            return (not s.empty and
                    s.map(lambda x: isinstance(x, (list, dict))).all())
        json_cols = [col for col in df.columns if is_json(col)]
        # merge with dtype from user
        new_dtype = {c:JSON for c in json_cols}
        if dtype is not None:
            new_dtype.update(dtype)
        new_dtype = None if new_dtype == {} else new_dtype
        # create sqlalchemy table model via pandas
        pandas_sql_engine = pd.io.sql.SQLDatabase(engine=engine, schema=schema)
        table = pd.io.sql.SQLTable(name=table_name,
                                   pandas_sql_engine=pandas_sql_engine,
                                   frame=df,
                                   dtype=new_dtype).table

        # change bindings of table (we want a sqlalchemy engine
        # not a pandas_sql_engine)
        metadata = MetaData(bind=engine)
        table.metadata = metadata

        # add PK
        constraint = PrimaryKeyConstraint(*[table.columns[name]
                                            for name in df.index.names])
        table.append_constraint(constraint)
        

        # add remaining attributes
        self.engine = engine
        self.df = df
        self.schema = schema
        self.table = table


    @staticmethod
    def _detect_db_type(engine) -> str:
        """
        Identifies whether the dialect of given sqlalchemy 
        engine corresponds to postgres, mysql or another sql type.

        Returns
        -------
        sql_type : {'postgres', 'mysql', 'sqlite', 'other'}
        """
        dialect = engine.dialect.dialect_description
        if re.search('psycopg|postgres', dialect):
            return "postgres"
        elif 'mysql' in dialect:
            return "mysql"
        elif 'sqlite' in dialect:
            return 'sqlite'
        else:
            return "other"
            
    def table_exists(self) -> bool:
        """
        Returns True if the table defined in given instance
        of PandasSpecialEngine exists else returns False.

        Returns
        -------
        exists : bool
            True if table exists else False
        """
        return self.engine.has_table(self.table.name, schema=self.schema)

    def create_schema_if_not_exists(self):
        """
        Creates the schema defined in given instance of
        PandasSpecialEngine if it does not exist.
        """
        if (self.schema is not None and
            not self.engine.dialect.has_schema(self.engine, self.schema)):
            self.engine.execute(CreateSchema(self.schema))

    def create_table_if_not_exists(self):
        """
        Creates the table generated in given instance of
        PandasSpecialEngine if it does not exist.
        """
        self.table.create(checkfirst=True)

    def get_db_columns_names(self) -> list:
        """
        Gets the column names of the SQL table defined
        in given instance of PandasSpecialEngine.

        Returns
        -------
        db_columns_names : list
            list of column names (str)
        """
        columns_info = self.engine.dialect.get_columns(connection=self.engine,
                                                       table_name=self.table.name,            
                                                       schema=self.schema)
        db_columns_names = [col_info["name"] for col_info in columns_info]
        return db_columns_names

    def add_new_columns(self):
        """
        Adds columns present in df but not in the SQL table
        for given instance of PandasSpecialEngine.

        Notes
        -----
        Sadly, it seems that we cannot create JSON columns.
        """
        # create deepcopies of the column because we are going to unbound
        # them for the table model (otherwise alembic would think we add
        # a column that already exists in the database)
        cols_to_add = [deepcopy(col) for col in self.table.columns
                       if col.name not in self.get_db_columns_names()]
        # check columns are not index levels
        if any((c.name in self.df.index.names for c in cols_to_add)):
            raise ValueError(('Cannot add any column that is part of the df index!\n'
                              "You'll have to update your table primary key or change your "
                              "df index"))
        
        with self.engine.connect() as con:
            ctx = MigrationContext.configure(con)
            op = Operations(ctx)
            for col in cols_to_add:
                col.table = None # Important! unbound column from table
                op.add_column(self.table.name, col, schema=self.schema)
                logger.info((f"Added column {col} (type: {col.type}) in table {self.table.name} "
                             f'(schema="{self.schema})"'))


    def get_db_table_schema(self):
        """
        Gets the sqlalchemy table model for the SQL table
        defined in given PandasSpecialEngine (using schema and
        table_name attributes to find the table in the database).

        Returns
        -------
        db_table : sqlalchemy.sql.schema.Table
        """
        table_name = self.table.name
        schema = self.schema
        engine = self.engine
        
        metadata = MetaData(bind=engine, schema=schema)
        metadata.reflect(bind=engine, schema=schema, only=[table_name])
        namespace = table_name if schema is None else f'{schema}.{table_name}'
        db_table = metadata.tables[namespace]
        return db_table

    def get_empty_columns(self) -> list:
        """
        Gets a list of the columns that contain no data
        in the SQL table defined in given instance of
        PandasSpecialEngine.
        Uses method get_db_table_schema (see its docstring).

        Returns
        -------
        empty_columns : list of str
            List of column names that contain no data
        """
        db_table = self.get_db_table_schema()
        empty_columns = []

        for col in db_table.columns:
            stmt = select(from_obj=db_table,
                          columns=[col],
                          whereclause=col.isnot(None)).limit(1)
            results = self.engine.execute(stmt).fetchall()
            if results == []:
                empty_columns.append(col)
        return empty_columns

    def adapt_dtype_of_empty_db_columns(self):
        """
        Changes the data types of empty columns in the SQL table defined
        in given instance of a PandasSpecialEngine.

        This should only happen in case of data type mismatches.
        This means with columns for which the sqlalchemy table
        model for df and the model for the SQL table have different data types.
        """
        empty_db_columns = self.get_empty_columns()
        db_table = self.get_db_table_schema()
        # if column does not have value in db and there are values
        # in the frame then change the column type if needed
        for col in empty_db_columns:
            # check if the column also exists in df
            if col.name not in self.df.columns:
                continue
            # check same type
            orig_type = db_table.columns[col.name].type.compile(self.engine.dialect)
            dest_type = self.table.columns[col.name].type.compile(self.engine.dialect)
            # remove character count e.g. "VARCHAR(50)" -> "VARCHAR" 
            orig_type = RE_CHARCOUNT_COL_TYPE.sub('', orig_type)
            dest_type = RE_CHARCOUNT_COL_TYPE.sub('', dest_type)
            # if same type or we want to insert TEXT instead of JSON continue
            # (JSON is not supported on some DBs so it's normal to have TEXT instead)
            if ((orig_type == dest_type) or
                ((orig_type == 'JSON') and (dest_type == 'TEXT'))):
                continue
            # grab the col/index from the df
            # so we can check if there are any values
            if col.name in self.df.index.names:
                df_col = self.df.index.get_level_values(col.name)
            else:
                df_col = self.df[col.name]
            if df_col.notna().any():
                # raise error if we have to modify the dtype but we have a SQlite engine
                # (SQLite does not support data type alteration)
                if self._db_type == 'sqlite':
                    raise ValueError('SQlite does not support column data type alteration!')
                with self.engine.connect() as con:
                    ctx = MigrationContext.configure(con)
                    op = Operations(ctx)
                    new_col = self.table.columns[col.name]
                    # check if postgres (in which case we have to use "using" syntax
                    # to alter columns data types)
                    if self._db_type == 'postgres':
                        escaped_col = str(new_col.compile(dialect=self.engine.dialect))
                        compiled_type = new_col.type.compile(dialect=self.engine.dialect)
                        alter_kwargs = {'postgresql_using':f'{escaped_col}::{compiled_type}'}
                    else:
                        alter_kwargs = {}
                    op.alter_column(table_name=self.table.name,
                                    column_name=new_col.name,
                                    type_=new_col.type,
                                    schema=self.schema,
                                    **alter_kwargs)
                    logger.info((f"Changed type of column {new_col.name} "
                                 f"from {col.type} to {new_col.type} "
                                 f'in table {self.table.name} (schema="{self.schema}")'))

    @staticmethod
    def _create_chunks(values, chunksize=10000):
        """
        Chunks a list into a list of lists of size
        :chunksize:.

        Parameters
        ----------
        chunksize : int > 0, default 10000
            Number of values to be inserted at once,
            an integer strictly above zero.
        """
        if not isinstance(chunksize, int) or chunksize <= 0:
            raise ValueError('chunksize must be an integer strictly above 0')
        chunks = [values[i:i + chunksize] for i in range(0, len(values), chunksize)]
        return chunks

    def _get_values_to_insert(self):
        """
        Gets the values to be inserted from the pandas DataFrame 
        defined in given instance of PandasSpecialEngine
        to the coresponding SQL table.

        Returns
        -------
        values : list
            Values from the df attribute that may have been converted
            for SQL compability e.g. pd.Timestamp will be converted
            to datetime.datetime objects.
        """
        # this seems to be the most reliable way to unpack
        # the DataFrame. For instance using df.to_dict(orient='records')
        # can introduce types such as numpy integer which we'd have to deal with
        values = self.df.reset_index().values.tolist()
        for i in range(len(values)):
            row = values[i]
            for j in range(len(row)):
                val = row[j]
                # replace pd.Timestamp with datetime.datetime
                if isinstance(val, pd.Timestamp):
                    values[i][j] = val.to_pydatetime()
                # check if na unless it is list like
                elif not pd.api.types.is_list_like(val) and pd.isna(val):
                    values[i][j] = null()
                # cast pd.Interval to str
                elif isinstance(val, pd.Interval):
                    warnings.warn(('found pd.Interval objects, '
                                   'they will be casted to str'))
                    values[i][j] = str(val)
        return values
    
    def upsert(self, if_row_exists, chunksize=10000):
        """
        Generates and executes an upsert (insert update or 
        insert ignore depending on :if_row_exists:) statement
        for given instance of PandasSpecialEngine.

        The values of df will be upserted with different sqlalchemy
        methods depending on the dialect (e.g. using
        sqlalchemy.dialects.postgresql.insert for postgres).
        See more information under pangres.upsert.

        Parameters
        ----------
        if_rows_exists : {'ignore', 'update'}
            If 'ignore' where the primary key matches nothing is
            updated.
            If 'update' where the primary key matches the values
            are updated using what's available in df.
            In both cases rows are inserted for non primary keys.
        chunksize : int > 0, default 900
            Number of values to be inserted at once,
            an integer strictly above zero.
        """
        # VERIFY ARGUMENTS
        if if_row_exists not in ('ignore', 'update'):
            raise ValueError('if_row_exists must be "ignore" or "update"')
        # convert values if needed
        values = self._get_values_to_insert()
        # recalculate chunksize for sqlite
        if self._db_type == 'sqlite':
            # to circumvent the max of 999 parameters for sqlite we have to
            # make sure chunksize is not too high also I don't know how to
            # deal with tables that have more than 999 columns because even
            # with single row inserts it's already too many variables.
            # (you can google "SQLITE_MAX_VARIABLE_NUMBER" for more info)
            new_chunksize = floor(999 / len(self.table.columns))
            if new_chunksize < 1:
                # case > 999 columns
                err = ('Updating SQlite tables with more than 999 columns is '
                       'not supported due to max variables restriction (999 max). '
                       'If you know a way around that please let me know'
                       '(e.g. post a GitHub issue)!')
                raise NotImpletementedError(err)
            if chunksize > new_chunksize:
                logger.warning((f'Reduced chunksize from {chunksize} to {new_chunksize} due '
                                'to SQlite max variable restriction (max 999).'))
            chunksize=new_chunksize
        # creat chunks
        chunks = self._create_chunks(values=values, chunksize=chunksize)
        upsert_funcs = {"postgres":postgres_upsert,
                        "mysql":mysql_upsert,
                        "sqlite":sqlite_upsert,
                        "other":sqlite_upsert}
        upsert_func = upsert_funcs[self._db_type]
        with self.engine.connect() as con:
            for chunk in chunks:
                upsert = upsert_func(engine=self.engine,
                                     connection=con,
                                     table=self.table,
                                     values=chunk,
                                     if_row_exists=if_row_exists)

    def __repr__(self):
        text = f"""PandasSpecialEngine (id {id(self)}, hexid {hex(id(self))})
                   * connection: {self.engine}
                   * schema: {self.schema}
                   * table: {self.table.name}
                   * SQLalchemy table model:\n{self.table.__repr__()}"""
        text = '\n'.join([line.strip() for line in text.splitlines()])
        
        df_repr = (str(self.df.head()) if not hasattr(self.df, 'to_markdown')
                   else str(self.df.head().to_markdown()))
        text += f'\n* df.head():\n{df_repr}'
        return text



#!/usr/bin/env python
# coding: utf-8
"""
Here we test different indices:
1. MultiIndex
2. Indices with NULL

# Note
Other problems with indices such as duplicated values are handled by pangres.helpersPandasSpecialEngine.
"""
import pandas as pd
from sqlalchemy import VARCHAR
from sqlalchemy.exc import IntegrityError
from pangres import upsert
from pangres.tests.conftest import drop_table_if_exists


# # Config

# +
# for creating table with MultiIndex
index_col = ['ix1', 'ix2', 'ix3']
df_multiindex = pd.DataFrame({'ix1': [1, 1], 'ix2': ['test', 'test2'], 
                              'ix3': [pd.Timestamp('2019-01-01'), pd.Timestamp('2019-01-02')],
                              'foo': [1, 2]}).set_index(index_col)

# for inserting values
df_multiindex2 = pd.DataFrame({'ix1': [2, 2], 'ix2': ['test', 'test2'], 
                              'ix3': [pd.Timestamp('2019-01-01'), pd.Timestamp('2019-01-02')],
                              'foo': [1, 2]}).set_index(index_col)

default_args = {'if_row_exists':'update',
                'adapt_dtype_of_empty_db_columns':False}


# -

# # Tests

# ## Test MultiIndex

def test_create_and_insert_table_multiindex(engine, schema):
    table_name = 'test_multiindex'
    namespace = f'{schema}.{table_name}' if schema is not None else table_name
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    # dtype for index for MySQL... (can't have flexible text length)
    dtype = {'ix2':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # create
    upsert(engine=engine, schema=schema, df=df_multiindex, table_name=table_name, dtype=dtype, **default_args)
    df_db = pd.read_sql(f'SELECT * FROM {namespace}', con=engine, index_col=index_col)

    # insert
    upsert(engine=engine, schema=schema, df=df_multiindex2, table_name=table_name, dtype=dtype, **default_args)
    df_db = pd.read_sql(f'SELECT * FROM {namespace}', con=engine, index_col=index_col)


# ## Test index with null value

def test_index_with_null(engine, schema):
    df = pd.DataFrame({'ix':[None], 'foo': [2]}).set_index('ix')
    table_name='test_index_with_null'
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    # don't test for mysql since only a warning is raised and the line is skipped
    if not 'mysql' in engine.dialect.dialect_description:
        try:
            upsert(engine=engine, schema=schema, df=df, table_name=table_name, **default_args)
            raise ValueError('upsert did not fail as expected with null value in index')
        except IntegrityError as e:
            print(f'upsert failed as expected with null value in index. Error was:\n\n{e}')



"""
This module tests that we can use a unique key instead of a primary key for upserting.

Thanks to LawrentChen on GitHub for pointing out this is possible and providing me
with an example.
See https://github.com/ThibTrip/pangres/issues/12
"""
import pandas as pd
import datetime
from pangres import upsert
from pangres.tests.conftest import drop_table_if_exists
from sqlalchemy import INTEGER, VARCHAR, MetaData, Column, UniqueConstraint
from sqlalchemy.ext.declarative import declarative_base

# # Helpers

# ## Table model

# +
table_name = 'test_unique_key'

def create_test_table(engine, schema):
    Base = declarative_base(bind=engine)

    class TestUniqueKey(Base):
        __tablename__ = table_name
        __table_args__ = (UniqueConstraint('order_id', 'product_id'),
                          {'schema':schema})

        row_id = Column(INTEGER, primary_key=True, autoincrement=True)
        order_id = Column(VARCHAR(5), nullable=False, server_default='-9999')
        product_id = Column(VARCHAR(5), nullable=False, server_default='-9999')
        qty = Column(INTEGER, nullable=True, comment='purchase_quantity')

    Base.metadata.create_all()


# -

# ## Data to be inserted/upserted in tests

# +
# initial data (first upsert)
data_old = {'order_id': ['A0001', 'A0002', 'A0002'],
            'product_id': ['PD100', 'PD200', 'PD201'],
            'qty': [10, 20, 22]}
df_old = pd.DataFrame(data_old).set_index(['order_id', 'product_id'])

# new data (second upsert)
data_new = {'order_id': ['A0001', 'A0002', 'A0002', 'A0003'],
            'product_id': ['PD100', 'PD200', 'PD201', 'PD300'],
            'qty': [10, 20, 77, 30]}
df_new = pd.DataFrame(data_new).set_index(['order_id', 'product_id'])


# -

# # Unique key test

def test_upsert_with_unique_keys(engine, schema):

    # helpers
    namespace = f'{schema}.{table_name}' if schema is not None else table_name
    read_from_db = lambda: pd.read_sql(f'SELECT * FROM {namespace}', con=engine, index_col='row_id')

    # create our test table
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    create_test_table(engine=engine, schema=schema)

    # add initial data (df_old)
    upsert(engine=engine, df=df_old, schema=schema, table_name='test_unique_key', if_row_exists='update')
    df = read_from_db()
    df_expected = df_old.assign(row_id=range(1, 4)).reset_index().set_index('row_id')
    pd.testing.assert_frame_equal(df, df_expected)

    # add new data (df_new)
    upsert(engine=engine, df=df_new, schema=schema, table_name='test_unique_key', if_row_exists='update')
    df = read_from_db()
    # before creating our expected df we need to implement the special case of postgres
    # where the id of the last row will be 7 instead of 4. I suppose that PG's ON
    # CONFLICT UPDATE clause will run in such a way that it will count 4 (number we
    # would expected) + 3 (three previous rows that were updated)
    last_row_id = 7 if 'postgres' in engine.dialect.dialect_description else 4
    df_expected = (pd.DataFrame([[1, 'A0001', 'PD100', 10],
                                 [2, 'A0002', 'PD200', 20],
                                 [3, 'A0002', 'PD201', 77],
                                 [last_row_id, 'A0003', 'PD300', 30]],
                                columns=['row_id'] + df_old.reset_index().columns.tolist())
                   .set_index('row_id'))
    pd.testing.assert_frame_equal(df, df_expected)
